@@ -5,10 +5,11 @@
 from __future__ import annotations
 
 import ctypes
+import sys
 import weakref
 from contextlib import contextmanager
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, List, Tuple, Union
+from typing import TYPE_CHECKING, Union
 from warnings import warn
 
 if TYPE_CHECKING:
@@ -28,6 +29,11 @@ _nvjitlink = None  # populated if nvJitLink can be used
 _nvjitlink_input_types = None  # populated if nvJitLink cannot be used
 
 
+def _nvjitlink_has_version_symbol(inner_nvjitlink) -> bool:
+    # This condition is equivalent to testing for version >= 12.3
+    return bool(inner_nvjitlink._inspect_function_pointer("__nvJitLinkVersion"))
+
+
 # Note: this function is reused in the tests
 def _decide_nvjitlink_or_driver() -> bool:
     """Returns True if falling back to the cuLink* driver APIs."""
@@ -37,28 +43,36 @@ def _decide_nvjitlink_or_driver() -> bool:
 
     _driver_ver = handle_return(driver.cuDriverGetVersion())
     _driver_ver = (_driver_ver // 1000, (_driver_ver % 1000) // 10)
-    try:
-        from cuda.bindings import nvjitlink as _nvjitlink
-        from cuda.bindings._internal import nvjitlink as inner_nvjitlink
-    except ImportError:
-        # binding is not available
-        _nvjitlink = None
-    else:
-        if inner_nvjitlink._inspect_function_pointer("__nvJitLinkVersion") == 0:
-            # binding is available, but nvJitLink is not installed
-            _nvjitlink = None
 
-    if _nvjitlink is None:
-        warn(
-            "nvJitLink is not installed or too old (<12.3). Therefore it is not usable "
-            "and the culink APIs will be used instead.",
-            stacklevel=3,
-            category=RuntimeWarning,
-        )
-        _driver = driver
-        return True
+    warn_txt_common = (
+        "the driver APIs will be used instead, which do not support"
+        " minor version compatibility or linking LTO IRs."
+        " For best results, consider upgrading to a recent version of"
+    )
+
+    try:
+        import cuda.bindings.nvjitlink as _nvjitlink
+    except ModuleNotFoundError:
+        warn_txt = f"cuda.bindings.nvjitlink is not available, therefore {warn_txt_common} cuda-bindings."
     else:
-        return False
+        from cuda.bindings._internal import nvjitlink as inner_nvjitlink
+
+        try:
+            if _nvjitlink_has_version_symbol(inner_nvjitlink):
+                return False  # Use nvjitlink
+        except RuntimeError:
+            warn_detail = "not available"
+        else:
+            warn_detail = "too old (<12.3)"
+        warn_txt = (
+            f"{'nvJitLink*.dll' if sys.platform == 'win32' else 'libnvJitLink.so*'} is {warn_detail}."
+            f" Therefore cuda.bindings.nvjitlink is not usable and {warn_txt_common} nvJitLink."
+        )
+        _nvjitlink = None
+
+    warn(warn_txt, stacklevel=2, category=RuntimeWarning)
+    _driver = driver
+    return True
 
 
 def _lazy_init():
@@ -140,14 +154,14 @@ class LinkerOptions:
     fma : bool, optional
         Use fast multiply-add.
         Default: True.
-    kernels_used : [Union[str, Tuple[str], List[str]]], optional
+    kernels_used : [Union[str, tuple[str], list[str]]], optional
         Pass a kernel or sequence of kernels that are used; any not in the list can be removed.
-    variables_used : [Union[str, Tuple[str], List[str]]], optional
+    variables_used : [Union[str, tuple[str], list[str]]], optional
         Pass a variable or sequence of variables that are used; any not in the list can be removed.
     optimize_unused_variables : bool, optional
         Assume that if a variable is not referenced in device code, it can be removed.
         Default: False.
-    ptxas_options : [Union[str, Tuple[str], List[str]]], optional
+    ptxas_options : [Union[str, tuple[str], list[str]]], optional
         Pass options to PTXAS.
     split_compile : int, optional
         Split compilation maximum thread count. Use 0 to use all available processors. Value of 1 disables split
@@ -177,10 +191,10 @@ class LinkerOptions:
     prec_div: bool | None = None
     prec_sqrt: bool | None = None
     fma: bool | None = None
-    kernels_used: Union[str, Tuple[str], List[str]] | None = None
-    variables_used: Union[str, Tuple[str], List[str]] | None = None
+    kernels_used: Union[str, tuple[str], list[str]] | None = None
+    variables_used: Union[str, tuple[str], list[str]] | None = None
     optimize_unused_variables: bool | None = None
-    ptxas_options: Union[str, Tuple[str], List[str]] | None = None
+    ptxas_options: Union[str, tuple[str], list[str]] | None = None
     split_compile: int | None = None
     split_compile_extended: int | None = None
     no_cache: bool | None = None
@@ -203,17 +217,17 @@ class LinkerOptions:
             self.formatted_options.append(f"-maxrregcount={self.max_register_count}")
         if self.time is not None:
             self.formatted_options.append("-time")
-        if self.verbose is not None:
+        if self.verbose:
             self.formatted_options.append("-verbose")
-        if self.link_time_optimization is not None:
+        if self.link_time_optimization:
             self.formatted_options.append("-lto")
-        if self.ptx is not None:
+        if self.ptx:
             self.formatted_options.append("-ptx")
         if self.optimization_level is not None:
             self.formatted_options.append(f"-O{self.optimization_level}")
-        if self.debug is not None:
+        if self.debug:
             self.formatted_options.append("-g")
-        if self.lineinfo is not None:
+        if self.lineinfo:
             self.formatted_options.append("-lineinfo")
         if self.ftz is not None:
             self.formatted_options.append(f"-ftz={'true' if self.ftz else 'false'}")
@@ -273,21 +287,21 @@ class LinkerOptions:
             self.option_keys.append(_driver.CUjit_option.CU_JIT_MAX_REGISTERS)
         if self.time is not None:
             raise ValueError("time option is not supported by the driver API")
-        if self.verbose is not None:
+        if self.verbose:
             self.formatted_options.append(1)
             self.option_keys.append(_driver.CUjit_option.CU_JIT_LOG_VERBOSE)
-        if self.link_time_optimization is not None:
+        if self.link_time_optimization:
             self.formatted_options.append(1)
             self.option_keys.append(_driver.CUjit_option.CU_JIT_LTO)
-        if self.ptx is not None:
+        if self.ptx:
             raise ValueError("ptx option is not supported by the driver API")
         if self.optimization_level is not None:
             self.formatted_options.append(self.optimization_level)
             self.option_keys.append(_driver.CUjit_option.CU_JIT_OPTIMIZATION_LEVEL)
-        if self.debug is not None:
+        if self.debug:
             self.formatted_options.append(1)
             self.option_keys.append(_driver.CUjit_option.CU_JIT_GENERATE_DEBUG_INFO)
-        if self.lineinfo is not None:
+        if self.lineinfo:
             self.formatted_options.append(1)
             self.option_keys.append(_driver.CUjit_option.CU_JIT_GENERATE_LINE_INFO)
         if self.ftz is not None:
@@ -395,10 +409,9 @@ class Linker:
 
     def _add_code_object(self, object_code: ObjectCode):
         data = object_code._module
-        assert_type(data, bytes)
         with _exception_manager(self):
             name_str = f"{object_code.name}"
-            if _nvjitlink:
+            if _nvjitlink and isinstance(data, bytes):
                 _nvjitlink.add_data(
                     self._mnff.handle,
                     self._input_type_from_code_type(object_code._code_type),
@@ -406,7 +419,13 @@ class Linker:
                     len(data),
                     name_str,
                 )
-            else:
+            elif _nvjitlink and isinstance(data, str):
+                _nvjitlink.add_file(
+                    self._mnff.handle,
+                    self._input_type_from_code_type(object_code._code_type),
+                    data,
+                )
+            elif (not _nvjitlink) and isinstance(data, bytes):
                 name_bytes = name_str.encode()
                 handle_return(
                     _driver.cuLinkAddData(
@@ -421,6 +440,21 @@ class Linker:
                     )
                 )
                 self._mnff.const_char_keep_alive.append(name_bytes)
+            elif (not _nvjitlink) and isinstance(data, str):
+                name_bytes = name_str.encode()
+                handle_return(
+                    _driver.cuLinkAddFile(
+                        self._mnff.handle,
+                        self._input_type_from_code_type(object_code._code_type),
+                        data.encode(),
+                        0,
+                        None,
+                        None,
+                    )
+                )
+                self._mnff.const_char_keep_alive.append(name_bytes)
+            else:
+                raise TypeError(f"Expected bytes or str, but got {type(data).__name__}")
 
     def link(self, target_type) -> ObjectCode:
         """
